using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;

namespace moyashi {
  public class ChromaKeyifier : MVRScript {
    private string fileRequestPath = null;
    private HashSet<MaterialReplacer> materialReplacers = new HashSet<MaterialReplacer>();
    private bool isEvenFrame = true;
    private bool isWireframeEnabled = false;
    private JSONStorableBool showWireframeStorable;
    private static string[] SUPPORTED_ATOM_TYPES = {
      "CustomUnityAsset",
      "Capsule",
      "Cube",
      "Sphere",
      "Slate",
      "Wall",
      "WoodPanel",
    };

    private string getPackagePath() {
      string filename = manager.GetJSON()["plugins"][name.Split('_')[0]].Value;
      int index = filename.IndexOf(":/");

      if (index > 0) {
        string path = filename.Substring(0, index + 2);
        // SuperController.LogMessage($"package path: {path}");
        return path;
      } else {
        return "";
      }
    }

    private IEnumerator loadShaders() {
      AssetLoadedMonitor assetLoadedMonitor = new AssetLoadedMonitor();
      AssetLoader.AssetBundleFromFileRequest fileRequest = new AssetLoader.AssetBundleFromFileRequest {
        path = getPackagePath() + $"Custom/Assets/moyashi/Moyashi_ChromaKeyifier2.assetbundle",
        callback = assetLoadedMonitor.finish
      };

      fileRequestPath = fileRequest.path;
      AssetLoader.QueueLoadAssetBundleFromFile(fileRequest);
      yield return StartCoroutine(assetLoadedMonitor);

      if (fileRequest.assetBundle != null) {
        Shader shader = fileRequest.assetBundle.LoadAsset<Shader>($"MoyashiChromaKeyifier2.shader");

        // SuperController.LogMessage($"containingAtom: {containingAtom.type} {containingAtom} {containingAtom.isSubSceneType}");
        // SuperController.LogMessage($"parents: {containingAtom.childAtomContainer}");

        if (SUPPORTED_ATOM_TYPES.Any(containingAtom.type.Contains)) { 
          List<Renderer> renderers = new List<Renderer>();
          // containingAtom.GetComponentsInChildren<Renderer>(false, renderers);
          containingAtom.childAtomContainer.gameObject.GetComponentsInChildren<Renderer>(false, renderers);

          // SuperController.LogMessage($"renderers: {renderers.Count} {renderers}");

          foreach(var renderer in renderers) {
            // SuperController.LogMessage($"{shader}");
            simpleAttachBehaviour(renderer, shader);
          }
        } else {
          SuperController.LogError($"Atom type {containingAtom.type} not supported.");
        }      
      } else {
        SuperController.LogError("assetbundle failed to load");
      }
    }

    private void simpleAttachBehaviour(Renderer renderer, Shader shader) {
      // SuperController.LogMessage("simple attach");
      MaterialReplacer materialReplacer = renderer.GetComponent<MaterialReplacer>();

      if (materialReplacer == null) {
        materialReplacer = renderer.gameObject.AddComponent<MaterialReplacer>();
        materialReplacer.shader = shader;
        materialReplacer.isWireframeEnabled = isWireframeEnabled;
      }

      materialReplacers.Add(materialReplacer);
    }

    // I'm sure there would be a better way of doing this...
    private void syncToCameraBackgroundColor() {
      var coreControl = GetSceneAtoms().First((atom) => atom.uid == "CoreControl");

      if (coreControl == null) return;
      // SuperController.LogMessage($"atom {coreControl.uid}");

      var pm = coreControl.gameObject.GetComponentsInChildren<MVRPluginManager>().First((x) => x.name == "ScenePluginManager");
      if (pm == null) return;

      // SuperController.LogMessage($"pm {pm.gameObject.transform.childCount}");

      for (int i = 0; i < pm.gameObject.transform.childCount; i++) {
        // SuperController.LogMessage($" - {pm.gameObject.transform.GetChild(i)}");
        var obj = pm.gameObject.transform.GetChild(i);
        var plugins = obj.GetComponents<MVRPlugin>();

        for (int j = 0; j < obj.transform.childCount; j++) {
          // SuperController.LogMessage($" - - {obj.transform.GetChild(j)}");

          var plugin = obj.transform.GetChild(j);
          if (plugin.name.Contains("CameraBackgroundColor")) {
            var storable = plugin.GetComponent<JSONStorable>();
            // SuperController.LogMessage("found it");
            
            if (storable != null) {
              var rgb = Shader.GetGlobalColor("_MoyashiChromaKeyifierGlobal_ChromaKey");
              var hsv = HSVColorPicker.RGBToHSV(rgb.r, rgb.g, rgb.b);
              storable.SetColorParamValue("Color", hsv);
            }

            return;
          }
        }
      }
    }

    public override void Init() {
      // SuperController.LogMessage("Init");

      if (!Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_INIT")) {
        // set some defaults if no init yet
        Shader.SetGlobalColor("_MoyashiChromaKeyifierGlobal_ChromaKey", new Color(0.392f, 0.471f, 0.392f));
        Shader.SetGlobalColor("_MoyashiChromaKeyifierGlobal_ShadowColor", new Color(0.0f, 0.0f, 0.0f));
        Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowFadeStrength", 0.3f);
        Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowEdgeRolloff", 0.3f);
        Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowFloor", 0.0f);

        Shader.EnableKeyword("MOYASHI_CHROMA_KEYIFIER_INIT");
      }

      {
        Color initialChromaKeyColor = Shader.GetGlobalColor("_MoyashiChromaKeyifierGlobal_ChromaKey");

        if (initialChromaKeyColor == null) {
          initialChromaKeyColor = new Color(0.392f, 0.471f, 0.392f);          
        }

        HSVColor hsvChromaKey = HSVColorPicker.RGBToHSV(0.392f, 0.471f, 0.392f);
        JSONStorableColor chromaKey = new JSONStorableColor("Chroma Key", hsvChromaKey);
        chromaKey.storeType = JSONStorableParam.StoreType.Full;
        var picker = CreateColorPicker(chromaKey, true);
        RegisterColor(chromaKey);
        chromaKey.setCallbackFunction = (float h, float s, float v) => { 
          Shader.SetGlobalColor("_MoyashiChromaKeyifierGlobal_ChromaKey", HSVColorPicker.HSVToRGB(h, s, v)); 

          if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_CHROMA_SYNC")) {
            syncToCameraBackgroundColor();
          }
        };
      }

      {
        JSONStorableBool storable = new JSONStorableBool("Sync Camera Background Color", false);
        storable.setCallbackFunction = (bool val) => {
          if(val) {
            Shader.EnableKeyword("MOYASHI_CHROMA_KEYIFIER_CHROMA_SYNC");
            syncToCameraBackgroundColor();
          } else {
            Shader.DisableKeyword("MOYASHI_CHROMA_KEYIFIER_CHROMA_SYNC");
          }
        };

        if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_CHROMA_SYNC")) {
          syncToCameraBackgroundColor();
          storable.valNoCallback = true;
        } else {
          storable.valNoCallback = false;
        }

        RegisterBool(storable);
        var toggle = CreateToggle(storable, true);
      }

      {
        Color initialShadowColor = Shader.GetGlobalColor("_MoyashiChromaKeyifierGlobal_ShadowColor");

        if (initialShadowColor == null) {
          initialShadowColor = new Color(0.0f, 0.0f, 0.0f);
        }

        HSVColor hsvShadowColor = HSVColorPicker.RGBToHSV(0.0f, 0.0f, 0.0f);
        JSONStorableColor shadowColor = new JSONStorableColor("Shadow Color", hsvShadowColor); // should this be JSONStorableVector3?
        shadowColor.storeType = JSONStorableParam.StoreType.Full;
        var picker = CreateColorPicker(shadowColor, false);
        RegisterColor(shadowColor);
        shadowColor.setCallbackFunction = (float h, float s, float v) => { Shader.SetGlobalColor("_MoyashiChromaKeyifierGlobal_ShadowColor", HSVColorPicker.HSVToRGB(h, s, v)); };
      }

      // {
      //   JSONStorableBool storable = new JSONStorableBool("Receive Shadows (TODO)", true);
      //   storable.setCallbackFunction = (bool val) => {
      //     // if(val) {
      //     //   Shader.EnableKeyword("MOYASHI_CHROMA_KEYIFIER_DITHER_ENABLED");
      //     // } else {
      //     //   Shader.DisableKeyword("MOYASHI_CHROMA_KEYIFIER_DITHER_ENABLED");
      //     // }
      //   };

      //   RegisterBool(storable);
      //   var toggle = CreateToggle(storable, false);
      // }

      {
        JSONStorableBool storable = new JSONStorableBool("Dither Shadows (checkerboard)", false);
        storable.setCallbackFunction = (bool val) => {
          if(val) {
            Shader.EnableKeyword("MOYASHI_CHROMA_KEYIFIER_DITHER_ENABLED");
          } else {
            Shader.DisableKeyword("MOYASHI_CHROMA_KEYIFIER_DITHER_ENABLED");
          }
        };

        if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_DITHER_ENABLED")) {
          storable.valNoCallback = true;
        } else {
          storable.valNoCallback = false;
        }

        RegisterBool(storable);
        var toggle = CreateToggle(storable, false);
      }

      {
        JSONStorableBool storable = new JSONStorableBool("Interleave Shadows", true);
        storable.setCallbackFunction = (bool val) => {
          if(val) {
            Shader.EnableKeyword("MOYASHI_CHROMA_KEYIFIER_INTERLEAVE_ENABLED");
          } else {
            Shader.DisableKeyword("MOYASHI_CHROMA_KEYIFIER_INTERLEAVE_ENABLED");
          }
        };

        if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_INTERLEAVE_ENABLED")) {
          storable.valNoCallback = true;
        } else {
          storable.valNoCallback = false;
        }

        RegisterBool(storable);
        var toggle = CreateToggle(storable, false);
      }

      {
        showWireframeStorable = new JSONStorableBool("Show Wireframe", false);
        showWireframeStorable.setCallbackFunction = (bool val) => {
          if(val) {
            Shader.EnableKeyword("MOYASHI_CHROMA_KEYIFIER_WIREFRAME_ENABLED");
            foreach(var materialReplacer in materialReplacers) {
              materialReplacer.EnableWireframePass(true);
            }
          } else {
            Shader.DisableKeyword("MOYASHI_CHROMA_KEYIFIER_WIREFRAME_ENABLED");
            foreach(var materialReplacer in materialReplacers) {
              materialReplacer.EnableWireframePass(false);
            }
          }
        };

        // init part
        if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_WIREFRAME_ENABLED")) {
          // isWireframeEnabled = true;
          showWireframeStorable.valNoCallback = true;
        } else {
          // isWireframeEnabled = false;
          showWireframeStorable.valNoCallback = false;
        }

        RegisterBool(showWireframeStorable);
        var toggle = CreateToggle(showWireframeStorable, false);
      }

      {
        float initialValue = 0.3f;
        float min = 0.0f;
        float max = 1.0f;
        JSONStorableFloat storable = new JSONStorableFloat("Shadow Fade Strength", initialValue, min, max, true, true);
        storable.storeType = JSONStorableParam.StoreType.Full;

        UIDynamicSlider slider = CreateSlider(storable, false);
        slider.valueFormat = "F3";
        slider.AutoSetButtons();

        RegisterFloat(storable);
        storable.setCallbackFunction = (float v) => { Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowFadeStrength", v); };
      }

      {
        float initialValue = 0.3f;
        float min = 0.0f;
        float max = 1.0f;
        JSONStorableFloat storable = new JSONStorableFloat("Shadow Edge Rolloff", initialValue, min, max, true, true);
        storable.storeType = JSONStorableParam.StoreType.Full;

        UIDynamicSlider slider = CreateSlider(storable, false);
        slider.valueFormat = "F3";
        slider.AutoSetButtons();

        RegisterFloat(storable);
        storable.setCallbackFunction = (float v) => { Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowEdgeRolloff", v); };
      }

      {
        float initialValue = 0.0f;
        float min = 0.0f;
        float max = 1.0f;
        JSONStorableFloat storable = new JSONStorableFloat("Shadow Noise Reduction", initialValue, min, max, true, true);
        storable.storeType = JSONStorableParam.StoreType.Full;

        UIDynamicSlider slider = CreateSlider(storable, false);
        slider.valueFormat = "F3";
        slider.AutoSetButtons();

        RegisterFloat(storable);
        storable.setCallbackFunction = (float v) => { Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowFloor", v); };
      }

      // {
      //   float initialValue = 0.0f;
      //   float min = 0.0f;
      //   float max = 1.0f;
      //   JSONStorableFloat storable = new JSONStorableFloat("Shadow Deadzone", initialValue, min, max, true, true);
      //   storable.storeType = JSONStorableParam.StoreType.Full;

      //   UIDynamicSlider slider = CreateSlider(storable, false);
      //   slider.valueFormat = "F3";
      //   slider.AutoSetButtons();

      //   RegisterFloat(storable);
      //   storable.setCallbackFunction = (float v) => { Shader.SetGlobalFloat("_MoyashiChromaKeyifierGlobal_ShadowDeadzone", v); };
      // }

      // hack fiesta
      // Defer(() => { 
      //   var rect = picker.colorPicker.Selector.rectTransform;
      //   var container = picker.colorPicker.Container;

      //   if (rect != null && container != null && container.rect != null) {
      //     rect.anchoredPosition = new Vector2(
      //       255 * picker.colorPicker.xVal,
      //       container.rect.height * picker.colorPicker.yVal
      //     );
      //   }
      // }, 0.0f);


      OnEnable(); // REEE
    }

    public void OnEnable() {
      // SuperController.LogMessage("OnEnable");
      if (containingAtom == null) return;

      // SuperController.LogMessage($"atom type: {containingAtom.type}");

      // SuperController.LogMessage("loading shaders...");
      StartCoroutine(loadShaders());


      // if (containingAtom.type == "Person") {
      //   StartCoroutine(attachBehaviour(containingAtom));
      // } else {
      //   SuperController.LogError("NostrilFix only works when attached to a Person");        
      // }
    }

    public void OnDisable() {
      // SuperController.LogMessage("disable");
      foreach(var materialReplacer in materialReplacers) {
        DestroyImmediate(materialReplacer);
      }

      materialReplacers.Clear();

      if (fileRequestPath != null) {
        AssetLoader.DoneWithAssetBundleFromFile(fileRequestPath);
        fileRequestPath = null;
      }
    }

    public void Update() {
      if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_INTERLEAVE_ENABLED")) {
        isEvenFrame = !isEvenFrame;
        Shader.SetGlobalInt("_MoyashiChromaKeyifierGlobal_EvenFrame", isEvenFrame ? 1 : 0);
      }

      if (!Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_WIREFRAME_ENABLED") && isWireframeEnabled) {
        isWireframeEnabled = false;
        showWireframeStorable.val = false;
      }

      if (Shader.IsKeywordEnabled("MOYASHI_CHROMA_KEYIFIER_WIREFRAME_ENABLED") && !isWireframeEnabled) {
        isWireframeEnabled = true;
        showWireframeStorable.val = true;
      }
    }
  }

  public class MaterialReplacer : MonoBehaviour {
    public Shader shader;
    public bool isWireframeEnabled;
    private Shader nostrilsOriginalShader;
    private Shader faceShader;
    private Material nostrilsOriginalMaterial;
    private Material nostrilsMaterial;
    // private Material faceMaterial;

    private Dictionary<Material, Material> materialCopies = new Dictionary<Material, Material>();

    public void EnableWireframePass(bool b) {
      isWireframeEnabled = b;
      foreach(KeyValuePair<Material, Material> entry in materialCopies) {
        entry.Key.SetShaderPassEnabled("Always", isWireframeEnabled);
      }
    }

    void Start() {
      // should already definitely have a renderer at this point
      var renderer = this.gameObject.GetComponent<Renderer>();

      // var materials = this.GetComponentsInChildren(typeof(Material));
      // var materials = this.GetComponents(typeof(Material));

      foreach(var material in renderer.materials) {
        // SuperController.LogMessage($"{material}");
        var backup = new Material(material.shader);
        backup.CopyPropertiesFromMaterial(material);
        materialCopies.Add(material, backup);

        // material.shader = Shader.Find("Unlit/Color");
        material.shader = shader;
        material.SetShaderPassEnabled("Always", isWireframeEnabled);
      }
    }

    void OnDestroy() {
      foreach(KeyValuePair<Material, Material> entry in materialCopies) {
        entry.Key.shader = entry.Value.shader;
        entry.Key.CopyPropertiesFromMaterial(entry.Value);
      }
    }
  }

  public class AssetLoadedMonitor : CustomYieldInstruction {
    private bool finished = false;

    public override bool keepWaiting { 
      get { return !finished; }
    }

    public void finish(AssetLoader.AssetBundleFromFileRequest fileRequest) {
      // SuperController.LogMessage("finish");
      this.finished = true;
      
      if (fileRequest.assetBundle == null) {
        AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      }
    }
  }
}